/*
 * Copyright (c) 2018 Michael Ko
 * 
 * This work is licensed under the GNU LGPLv3 license.
 * <https://www.gnu.org/licenses/lgpl-3.0.en.html>.
 */

/* global define, Function, global */

(function (root, factory) {
    // UMD (Universal Module Definition) https://github.com/umdjs/umd
    if (typeof define === "function" && define.amd) {
        define(["go-types"], factory);
    } else if (typeof exports === "object") {
        module.exports = factory(require("go-types"));
    } else {
        root.expect = factory(root.Types);
    }

}(typeof window !== "undefined" ? window : global, function (Types) {

    /**
     * Type-checking utilities
     * @author Michael Ko <koyote130708@gmail.com>
     * @version 1.0.0
     */

    var DEFAULT_TEMPLATE = "Expected ${name} to be ${expected}, but found: ${actual}";
    var DEFAULT_NAME = "the value";

    var helpers = {
        assert: function (expression, errorMsg, expected, actual, name, errorType) {
            if (!expression) {
                var template = errorMsg !== undefined ? errorMsg : DEFAULT_TEMPLATE;
                var msg = helpers.inject(template, {name: name !== undefined ? name : "the value", expected: JSON.stringify(expected), actual: JSON.stringify(actual)});
                var ErrorType = errorType !== undefined ? errorType : Error;

                throw new ErrorType(msg);
            }
        },
        inject: function (str, data) {
            var result = "";
            var offset = 0;
            var regex = new RegExp("\\${.+?}", "g");

            var match;

            while ((match = regex.exec(str))) {
                var matched = match[0];
                var index = match.index;
                var param = matched.substring(2, matched.length - 1);
                var p = data[param];
                var value = p;
                result += str.substring(offset, index) + value;
                offset = index + matched.length;
            }

            if (offset < str.length) {
                result += str.substring(offset, str.length);
            }
            return result;
        },
        error: function (expected, actual, name, template, errorType) {
            template = template !== undefined ? template : DEFAULT_TEMPLATE;
            var msg = helpers.inject(template, {name: name !== undefined ? name : DEFAULT_NAME, expected: expected, actual: actual});
            var ErrorType = errorType !== undefined ? errorType : Error;
            return new ErrorType(msg);
        },
        isEqual: function (a, b) {
            if (a instanceof Object) {
                if (Types.isFunction(a.equals) && !a.equals(b)) {
                    return false;
                }
                if (b instanceof Object) {
                    if (Types.isFunction(b.equals) && !b.equals(a)) {
                        return false;
                    }
                    if (Types.isFunction(a) && Types.isFunction(b)) {
                        return a === b;
                    }
                    if (a.constructor !== b.constructor) {
                        return false;
                    }
                    var props1 = Object.getOwnPropertyNames(a).sort();
                    var props2 = Object.getOwnPropertyNames(b).sort();

                    if (props1.length !== props2.length) {
                        return false;
                    }

                    var p1, p2;
                    for (var j = 0; j < props1.length; j++) {
                        p1 = props1[j];
                        p2 = props2[j];
                        if (!(p1 === p2 && helpers.isEqual(a[p1], b[p2]))) {
                            return false;
                        }
                    }
                } else {
                    return false;
                }
            } else if (a !== b) {
                return false;
            }
            return true;
        }
    };


    /**
     * Checks if the value is true and throws an error if not.
     * @param {*} value The value to check.
     * @param {*} expected The expected value
     * @param {string} [name] The name of the value
     * @param {string} [errorMsg] A custom message to describe the error.
     * @param {Error} [errorType] The type of the error to throw
     * @returns {*} the value passed.
     * @throws {TypeError} if the value is undefined.
     * @example <caption>Valid inputs</caption>
     * expect.true(true);
     * @example <caption>Invalid inputs</caption>
     * expect.true(false); 
     * // => TypeError: Expected the value to be true, but found: false
     * 
     * expect.true(false, "canFly");
     * // => TypeError: Expected canFly to be true, but found: false
     * 
     * expect.true(false, "canFly", "${name} should be ${expected}, but found: ${actual}");
     * // => TypeError: canFly should be true, but found: false
     * @since 1.0.0
     * @static
     */
    function expect(value, expected, name, errorMsg, errorType) {
        helpers.assert(helpers.isEqual(value, expected), errorMsg, expected, value, name, errorType);
        return value;
    }

    Object.assign(expect, {
        /**
         * Checks if the value is true and throws an error if not.
         * @param {*} value The value to check.
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] A custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * expect.true(true);
         * @example <caption>Invalid inputs</caption>
         * expect.true(false); 
         * // => TypeError: Expected the value to be true, but found: false
         * 
         * expect.true(false, "canFly");
         * // => TypeError: Expected canFly to be true, but found: false
         * 
         * expect.true(false, "canFly", "${name} should be ${expected}, but found: ${actual}");
         * // => TypeError: canFly should be true, but found: false
         * @since 1.0.0
         * @static
         */
        true: function (value, name, errorMsg) {
            helpers.assert(value === true, errorMsg, true, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is false.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        false: function (value, name, errorMsg) {
            helpers.assert(value === false, errorMsg, false, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @see Types.isUndefined
         * @since 1.0.0
         * @static
         */
        undefined: function (value, name, errorMsg) {
            helpers.assert(value === undefined, errorMsg, undefined, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        null: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, null, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a boolean.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a primitive boolean value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        boolean: function (value, name, errorMsg) {
            helpers.assert(Types.isBoolean(value), errorMsg, "boolean", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a number.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a primitive number.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        number: function (value, name, errorMsg) {
            helpers.assert(Types.isNumber(value), errorMsg, "number", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a string.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a primitive string.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        string: function (value, name, errorMsg) {
            helpers.assert(Types.isString(value), errorMsg, "string", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a primitive type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a primitive type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        primitive: function (value, name, errorMsg) {
            helpers.assert(Types.isPrimitive(value), errorMsg, "primitive", value, name, TypeError);
            return value;
        },
        /**
         * Ensues the value is an object.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an instance of Object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        object: function (value, name, errorMsg) {
            helpers.assert(Types.isObject(value), errorMsg, "object", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a plain object.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a plain object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        plainObject: function (value, name, errorMsg) {
            helpers.assert(Types.isPlainObject(value), errorMsg, "plain object", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a function.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an instance of Function.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        function: function (value, name, errorMsg) {
            helpers.assert(Types.isFunction(value), errorMsg, "function", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an array.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an instance of Array.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        array: function (value, name, errorMsg) {
            helpers.assert(Types.isArray(value), errorMsg, "array", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an array-like object.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an array-like object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        arrayLike: function (value, name, errorMsg) {
            helpers.assert(Types.isArrayLike(value), errorMsg, "array-like", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an array.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {(string|Function)} type a custom message to describe the error.
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an instance of Array.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        arrayOf: function (value, type, name, errorMsg) {
            helpers.assert(Types.isArrayOf(value, type), errorMsg, "array of " + type, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an array.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {(string|Function)} type a custom message to describe the error.
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an instance of Array.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        arrayLikeOf: function (value, type, name, errorMsg) {
            helpers.assert(Types.isArrayLikeOf(value, type), errorMsg, "array of " + type, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        date: function (value, name, errorMsg) {
            helpers.assert(Types.isDate(value), errorMsg, "Date", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        regExp: function (value, name, errorMsg) {
            helpers.assert(Types.isRegExp(value), errorMsg, "RegExp", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        promise: function (value, name, errorMsg) {
            helpers.assert(Types.isPromise(value), errorMsg, "Promise", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        promiseLike: function (value, name, errorMsg) {
            helpers.assert(Types.isPromiseLike(value), errorMsg, "Promise-like", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a thenable object.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        thenable: function (value, name, errorMsg) {
            helpers.assert(Types.isThenable(value), errorMsg, "thenable object", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a Symbol.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a Symbol.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        symbol: function (value, name, errorMsg) {
            helpers.assert(Types.isSymbol(value), errorMsg, "Symbol", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        bigInt: function (value, name, errorMsg) {
            helpers.assert(Types.isBigInt(value), errorMsg, "BigInt", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        map: function (value, name, errorMsg) {
            helpers.assert(Types.isMap(value), errorMsg, "Map", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        set: function (value, name, errorMsg) {
            helpers.assert(Types.isSet(value), errorMsg, "Set", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        weakMap: function (value, name, errorMsg) {
            helpers.assert(Types.isWeakMap(value), errorMsg, "WeakMap", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a BigInt.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a BigInt.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        weakSet: function (value, name, errorMsg) {
            helpers.assert(Types.isWeakSet(value), errorMsg, "WeakSet", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an iterable.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an array-like object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        iterable: function (value, name, errorMsg) {
            helpers.assert(Types.isIterable(value), errorMsg, "iterable", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an iterable.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an array-like object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        error: function (value, name, errorMsg) {
            helpers.assert(Types.isError(value), errorMsg, "Error", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an iterable.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an array-like object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        node: function (value, name, errorMsg) {
            helpers.assert(Types.isNode(value), errorMsg, "Node", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an iterable.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an array-like object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        element: function (value, name, errorMsg) {
            helpers.assert(Types.isElement(value), errorMsg, "Element", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an iterable.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not an array-like object.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        fragment: function (value, name, errorMsg) {
            helpers.assert(Types.isFragment(value), errorMsg, "DocumentFragment", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not null or undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is null or undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        nullish: function (value, name, errorMsg) {
            helpers.assert(Types.isNullish(value), errorMsg, "nullish", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a truthy value. All values are truthy unless they are defined as falsy (i.e., except for false, 0, 0n, "", null, undefined, and NaN).
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {Error} if the value is not a truthy value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        truthy: function (value, name, errorMsg) {
            helpers.assert(Types.isTruthy(value), errorMsg, "truthy", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a falsy value. There are 7 falsy values in JavaScript: false, 0, 0n, "", null, undefined, and NaN.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {Error} if the value is not a falsy value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        falsy: function (value, name, errorMsg) {
            helpers.assert(Types.isFalsy(value), errorMsg, "falsy", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not null or undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is null or undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        empty: function (value, name, errorMsg) {
            helpers.assert(Types.isEmpty(value), errorMsg, "empty", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not null or undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is null or undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        blank: function (value, name, errorMsg) {
            helpers.assert(Types.isBlank(value), errorMsg, "blank", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a numeric value.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a numeric value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        numeric: function (value, name, errorMsg) {
            helpers.assert(Types.isNumeric(value), errorMsg, "numeric", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a numeric value.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a numeric value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        integer: function (value, name, errorMsg) {
            helpers.assert(Types.isInteger(value), errorMsg, "integer", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a numeric value.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a numeric value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        safeInteger: function (value, name, errorMsg) {
            helpers.assert(Types.isSafeInteger(value), errorMsg, "safe integer", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a numeric value.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a numeric value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        float: function (value, name, errorMsg) {
            helpers.assert(Types.isFloat(value), errorMsg, "float", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a numeric value.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a numeric value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        finite: function (value, name, errorMsg) {
            helpers.assert(Types.isFinite(value), errorMsg, "finite", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is an infinite value.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is not a numeric value.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        infinite: function (value, name, errorMsg) {
            helpers.assert(Types.isInfinite(value), errorMsg, "infinite", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is NaN.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is NaN.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        NaN: function (value, name, errorMsg) {
            helpers.assert(Types.isNaN(value), errorMsg, "NaN", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {(string|Function|Array|null|undefined)} type
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        typeOf: function (value, type, name, errorMsg) {
            var expected = type;
            if (Types.isArray(type)) {
                expected = type.length > 1 ? type.slice(0, type.length - 1).join(", ") + " or " + type[type.length - 1] : type.join("");
            }
            helpers.assert(Types.isTypeOf(value, type), errorMsg, expected, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a native function.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {Error} if the value is a non-native function or is not a function.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        native: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "native function", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        zero: function (value, name, errorMsg) {
            helpers.assert(value === 0, errorMsg, 0, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        even: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "even number", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        odd: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "odd number", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        positive: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "positive number", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        negative: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "negative number", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        inRange: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "in range", value, name, RangeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        lower: function (value, high, name, errorMsg) {
            helpers.assert(value < high, errorMsg, "lower than " + high, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        higher: function (value, low, name, errorMsg) {
            helpers.assert(value > low, errorMsg, "higher than " + low, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {*} expected the expected value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        equal: function (value, expected, name, errorMsg) {
            helpers.assert(helpers.isEqual(value, expected), errorMsg, expected, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {*} expected the expected value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        same: function (value, expected, name, errorMsg) {
            helpers.assert(value === expected, errorMsg, expected, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {*} expected the expected value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        property: function (value, prop, name, errorMsg) {
            helpers.assert(value instanceof Object && value[prop], errorMsg, prop, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value has the specified type.
         * @param {*} value
         * @param {*} expected the expected value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg]
         * @returns {*} the value passed.
         * @throws {TypeError} if the value does not have the specified type.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        properties: function (value, props, propType, name, errorMsg) {
            helpers.assert(value instanceof Object && Types.isTypeOf(value[props], propType), errorMsg, propType, value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        match: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        length: function (value, expected, name, errorMsg) {
            helpers.assert(value instanceof Object && value.length === expected, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        lengthRange: function (value, atLeast, atMost, name, errorMsg) {
            helpers.assert(value instanceof Object && value.length >= atLeast && value.length <= atMost, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        pair: function (value, name, errorMsg) {
            helpers.assert(Types.isArray(value) && value.length === 2, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        keyValuePair: function (value, name, errorMsg) {
            helpers.assert(Types.isArray(value) && value.length === 2, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        alphabetic: function (value, name, errorMsg) {
            helpers.assert(Types.isArray(value) && value.length === 2, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        alphanumeric: function (value, name, errorMsg) {
            helpers.assert(Types.isArray(value) && value.length === 2, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not undefined.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is undefined.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        notUndefined: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "not undefined", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is not null.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is null.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        notNull: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "not null", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is NaN.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {TypeError} if the value is NaN.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        notNaN: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "not NaN", value, name, TypeError);
            return value;
        },
        /**
         * Checks the value is a non-native function.
         * @param {*} value
         * @param {string} [name] The name of the value
         * @param {string} [errorMsg] a custom message to describe the error.
         * @returns {*} the value passed.
         * @throws {Error} if the value is a native function or is not a function.
         * @example <caption>Valid inputs</caption>
         * // no error
         * expect.boolean(true);
         * @example <caption>Invalid inputs</caption>
         * // throws an error
         * expect.boolean(1);
         * @since 1.0.0
         * @static
         */
        notNativeFunction: function (value, name, errorMsg) {
            helpers.assert(value === null, errorMsg, "not native function", value, name, TypeError);
            return value;
        }
    });

    return expect;
}));