/*
 * Copyright (c) 2018 Michael Ko
 * 
 * This work is licensed under the GNU LGPLv3 license.
 * <https://www.gnu.org/licenses/lgpl-3.0.en.html>.
 */

/* global BigInt, Symbol */

"use strict";

var assert = require("chai").assert;

var Expect = require("../src/Expect");

suite("go/Expect", function () {
    var BIGINT = typeof BigInt !== "undefined" ? BigInt(9876543210123456789) : null;
    var SYMBOL = typeof Symbol !== "undefined" ? Symbol("abc") : null;
    var ARRAY_LIKE = {length: 0};
    var ITERABLE = {};
    var FN = function () {};

    if (SYMBOL) {
        ITERABLE[Symbol.iterator] = FN;
    }

    setup(function () {
    });

    suite("#truthy", function () {
        var truthy = Expect.truthy;

        test("normal", function () {
            truthy(true);
            truthy(1);
            truthy("false");
            truthy([]);
            truthy({});
        });

        test("exception", function () {
            assert.throws(truthy, false);
            assert.throws(truthy, null);
            assert.throws(truthy, undefined);
            assert.throws(truthy, "");
            assert.throws(truthy, 0);
            assert.throws(truthy, Number.NaN);
        });
    });

    suite("#falsy", function () {
        var falsy = Expect.falsy;

        test("normal", function () {
            falsy(undefined);
            falsy(null);
            falsy(false);
            falsy(0);
            falsy("");
            falsy(Number.NaN);
        });

        test("exception", function () {
            assert.throws(falsy, true);
            assert.throws(falsy, "false");
            assert.throws(falsy, "0");
        });
    });

    suite("#notUndefined", function () {
        var notUndefined = Expect.notUndefined;

        test("normal", function () {
            notUndefined(null);
            notUndefined(false);
            notUndefined(0);
            notUndefined("");
            notUndefined([]);
            notUndefined({});
        });

        test("exception", function () {
            assert.throws(notUndefined);
            assert.throws(notUndefined, undefined);
        });
    });


    suite("#notNull", function () {
        var notNull = Expect.notNull;

        test("normal", function () {
            notNull(undefined);
            notNull(false);
            notNull(0);
            notNull("");
            notNull([]);
            notNull({});
        });

        test("exception", function () {
            assert.throws(notNull, null);
        });
    });

    suite("#notNullish", function () {
        var notNullish = Expect.notNullish;

        test("normal", function () {
            notNullish(false);
            notNullish(0);
            notNullish("");
            notNullish([]);
            notNullish({});
        });

        test("exception", function () {
            assert.throws(notNullish);
            assert.throws(notNullish, undefined);
            assert.throws(notNullish, null);
        });
    });


    suite("#string", function () {
        var string = Expect.string;

        test("normal", function () {
            string("");
            string("1");
            string(new String(""));
        });

        test("exception", function () {
            assert.throws(string, 1);
            assert.throws(string, []);
        });
    });

    suite("#number", function () {
        var number = Expect.number;

        test("normal", function () {
            number(0);
            number(0.000001);
            number(Number.NEGATIVE_INFINITY);
            number(Number.POSITIVE_INFINITY);
            number(new Number(1));
            number(Number.NaN);
        });

        test("exception", function () {
            assert.throws(number, "1");
            assert.throws(number, BIGINT);
        });
    });

    suite("#boolean", function () {
        var boolean = Expect.boolean;

        test("normal", function () {
            boolean(false);
            boolean(true);
            boolean(new Boolean(true));
        });

        test("exception", function () {
            assert.throws(boolean, 0);
            assert.throws(boolean, null);
            assert.throws(boolean, "true");
        });
    });

    suite("#numeric", function () {
        var numeric = Expect.numeric;

        test("normal", function () {
            numeric(1);
            numeric(Number.POSITIVE_INFINITY);
            numeric(Number.NEGATIVE_INFINITY);
            numeric(Number.NaN);
            numeric("1");
            numeric("-1.2345E6");

            if (BIGINT) {
                numeric(BIGINT);
            }
        });

        test("exception", function () {
            assert.throws(numeric, "");
            assert.throws(numeric, "1a");
        });
    });

    suite("#primitive", function () {
        var primitive = Expect.primitive;

        test("normal", function () {
            primitive(undefined);
            primitive(null);
            primitive(false);
            primitive(0);
            primitive("");

            if (SYMBOL) {
                primitive(SYMBOL);
            }

            if (BIGINT) {
                primitive(BIGINT);
            }
        });

        test("exception", function () {
            assert.throws(primitive, {});
            assert.throws(primitive, []);
            assert.throws(primitive, FN);
        });
    });

    suite("#object", function () {
        var object = Expect.object;

        test("normal", function () {
            object({});
            object([]);
            object(FN);
        });

        test("exception", function () {
            assert.throws(object, null);
            assert.throws(object, 1);
            assert.throws(object, "");
        });
    });

    suite("#function", function () {
        var expectFn = Expect.function;

        test("normal", function () {
            expectFn(FN);
        });

        test("exception", function () {
            assert.throws(expectFn, null);
            assert.throws(expectFn, new FN());
        });
    });

    suite("#array", function () {
        var array = Expect.array;

        test("normal", function () {
            array([]);
        });

        test("exception", function () {
            assert.throws(array, {});
            assert.throws(array, "");
            assert.throws(array, FN);
            assert.throws(array, ARRAY_LIKE);
        });
    });

    suite("#arrayLike", function () {
        var arrayLike = Expect.arrayLike;

        test("normal", function () {
            arrayLike([]);
            arrayLike(ARRAY_LIKE);
        });

        test("exception", function () {
            assert.throws(arrayLike, "");
            assert.throws(arrayLike, FN);
        });
    });

    suite("#iterable", function () {
        var iterable = Expect.iterable;

        test("normal", function () {
            iterable("");
            iterable([]);
            iterable(ITERABLE);
        });

        test("exception", function () {
            assert.throws(iterable, FN);
            assert.throws(iterable, ARRAY_LIKE);
        });
    });

    suite("#plainObject", function () {
        var plainObject = Expect.plainObject;

        test("normal", function () {
            plainObject({});
        });

        test("exception", function () {
            assert.throws(plainObject, null);
            assert.throws(plainObject, []);
            assert.throws(plainObject, FN);
            assert.throws(plainObject, new FN());
        });
    });

    suite("#type", function () {
        var type = Expect.type;

        test("normal", function () {
            type(null, "object");
            type(undefined, "undefined");
            type(false, "boolean");
            type(0, "number");
            type("", "string");
            type(FN, "function");
        });

        test("exception", function () {
            assert.throws(type, 0, "boolean");
        });
    });


    suite("#native", function () {
        var native = Expect.native;

        test("normal", function () {
            native(log);
        });

        test("exception", function () {
            assert.throws(native, fn1);
            assert.throws(native, fn2);
            assert.throws(native, new fn1());
            assert.throws(native, null);
        });

    });


    suite("#nonNative", function () {
        var nonNative = Expect.nonNative;

        test("normal", function () {
            nonNative(fn1);
            nonNative(fn2);
        });

        test("exception", function () {
            assert.throws(nonNative, log);
            assert.throws(nonNative, new fn1());
            assert.throws(nonNative, null);
        });

    });

});
